/*
 *  Render.cpp
 *  Tcc
 */

#include "Render.h"
#include "vtkmsqAnalyzeReader.h"
#include "vtkmsqImageProperties.h"
#include <stdlib.h>
#include <QObject>
#include <QFileDialog>
#include <QMessageBox>
#include <QTableWidgetItem>

#include <vtkColorTransferFunction.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkInteractorStyleTrackballCamera.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkPiecewiseFunction.h>
#include <vtkVolumeProperty.h>
#include <vtkVolumeRayCastMIPFunction.h>
#include <vtkVolumeRayCastMapper.h>
#include <vtkVolumeMapper.h>
#include <vtkVolume.h>
#include <vtkCommand.h>
#include <vtkBoxWidget.h>
#include <vtkTransform.h>
#include <vtkEventQtSlotConnect.h>
#include <vtkSmartPointer.h>
#include <vtkImageReader.h>
#include <vtkImageData.h>
#include <vtkPointData.h>
#include <vtkLookupTable.h>

using namespace std;

Render::Render()
{
	this->setupUi(this);
	this->properties = vtkmsqImageProperties::New();
	
	action_Open->setShortcuts(QKeySequence::Open);
	action_Open->setStatusTip(tr("Open an existing image file"));
	action_Open->setIconVisibleInMenu(false);
	connect(action_Open, SIGNAL(triggered()), this, SLOT(fileOpen()));

	action_Quit->setShortcuts(QKeySequence::Quit);
	action_Quit->setStatusTip(tr("Quit the program"));
	action_Quit->setIconVisibleInMenu(false);
	connect(action_Quit, SIGNAL(triggered()), qApp, SLOT(quit()));
	disableButtons();
	connect(CreateRGBA, SIGNAL(clicked()), this, SLOT(AddRGBAPoint()));
	connect(RemoveRGBA, SIGNAL(clicked()), this, SLOT(RemoveRGBAPoint()));
	connect(RGBATableWidget, SIGNAL(itemClicked(QTableWidgetItem *)), this, SLOT(updateRGBA()));
	connect(WhiteButton, SIGNAL(clicked()), this, SLOT(changeToWhite()));
	connect(BlackButton, SIGNAL(clicked()), this, SLOT(changeToBlack()));
}

void Render::changeToBlack()
{
	renderer->SetBackground(0.0, 0.0, 0.0);
	declarativeView->update();
}

void Render::changeToWhite()
{
	renderer->SetBackground(1.0, 1.0, 1.0);
	declarativeView->update();
}

void Render::updateRGBA()
{
	int row = RGBATableWidget->currentRow();
	
	RGBASpinBoxIntensity->setValue(RGBATableWidget->item(row, 0)->text().toInt());
	SpinBoxRed->setValue(RGBATableWidget->item(row, 1)->text().toInt());
	SpinBoxGreen->setValue(RGBATableWidget->item(row, 2)->text().toInt());
	SpinBoxBlue->setValue(RGBATableWidget->item(row, 3)->text().toInt());
	SpinBoxAlpha->setValue(RGBATableWidget->item(row, 4)->text().toInt());
}

void Render::disableButtons()
{
	RGBASpinBoxIntensity->setEnabled(false);
	RGBASliderIntensity->setEnabled(false);
	SliderRed->setEnabled(false);
	SliderGreen->setEnabled(false);
	SliderBlue->setEnabled(false);
	SliderAlpha->setEnabled(false);

	SpinBoxRed->setEnabled(false);
	SpinBoxGreen->setEnabled(false);
	SpinBoxBlue->setEnabled(false);
	SpinBoxAlpha->setEnabled(false);
	CreateRGBA->setEnabled(false);
	RemoveRGBA->setEnabled(false);
	BlackButton->setEnabled(false);
	WhiteButton->setEnabled(false);
}

void Render::setCurrentFile(const QString &fileName)
{
  currentFileName = fileName;
  setWindowFilePath(fileName);
	declarativeView->update();
}

void Render::AddRGBAPoint()
{
	int intensity, red, green, blue, alpha;
	red = SpinBoxRed->value();
	green = SpinBoxGreen->value();
	blue = SpinBoxBlue->value();
	intensity = RGBASpinBoxIntensity->value();
	alpha = SpinBoxAlpha->value();

	colorTransferFunction->AddRGBPoint(intensity, (red/255.0), (green/255.0), (blue/255.0));
	opacityTransferFunction->AddPoint(intensity, (alpha/100.0));

	insertRow(intensity, red, green, blue, alpha);

	declarativeView->update();
}

void Render::RemoveRGBAPoint()
{
	int intensity = RGBASpinBoxIntensity->value();

	colorTransferFunction->RemovePoint(intensity);
	opacityTransferFunction->RemovePoint(intensity);

	for(int i = 0; i < RGBATableWidget->rowCount(); i++) {
		if(RGBATableWidget->item(i, 0)->text().toInt() == intensity) {
			RGBATableWidget->removeRow(i);
		}
	}
	declarativeView->update();
}

void Render::fileOpen() {
	QString fileName = QFileDialog::getOpenFileName(this, tr("Open Image"),
			currentFileName,
			tr("Analyze (*.hdr *.img)"),
			&currentFilter);

	int success = 1;

	if (!fileName.isEmpty()) {
		if (currentFilter == "Analyze (*.hdr *.img)") {
			disableButtons();
			success = loadAnalyzeImage(fileName.toLocal8Bit().constData());
		}	else {
			success = 0;
		}

		if (success) {
			setCurrentFile(fileName);
		}
		else {
			warningMessage(tr("Error reading file %1.").arg(fileName),
					tr("Make sure file is of correct type and retry."));
		}
	}
}

void Render::warningMessage(const QString &text, const QString &info)
{
	QMessageBox msgBox;
	msgBox.setText(text);
	msgBox.setInformativeText(info);
	msgBox.setIcon(QMessageBox::Warning);
	msgBox.exec();
}

void Render::configureButtons(int min, int max) 
{
	int window = (max - min)/3.0;

	RGBASliderIntensity->setMinimum(min);
	RGBASliderIntensity->setMaximum(max);
	RGBASliderIntensity->setEnabled(true);

	RGBASpinBoxIntensity->setMinimum(min);
	RGBASpinBoxIntensity->setMaximum(max);
	RGBASpinBoxIntensity->setEnabled(true);

	RGBASliderIntensity->setValue(window);

	SliderRed->setEnabled(true);
	SliderRed->setValue(0);
	SliderGreen->setEnabled(true);
	SliderGreen->setValue(0);
	SliderBlue->setEnabled(true);
	SliderBlue->setValue(0);
	SliderAlpha->setEnabled(true);
	SliderAlpha->setValue(0);

	SpinBoxRed->setEnabled(true);
	SpinBoxGreen->setEnabled(true);
	SpinBoxBlue->setEnabled(true);
	SpinBoxAlpha->setEnabled(true);
	CreateRGBA->setEnabled(true);
	RemoveRGBA->setEnabled(true);

	for(int i =  RGBATableWidget->rowCount(); i >= 0; i--) {
		RGBATableWidget->removeRow(i);
	}

	BlackButton->setEnabled(true);
	WhiteButton->setEnabled(true);
}

void Render::insertRow(int intensity, int R, int G, int B, int A)
{
	int row = -1;
	int numberOfRows = RGBATableWidget->rowCount();
	for(int i = 0; i < numberOfRows; i++){
		if(RGBATableWidget->item(i, 0)->text().toInt() == intensity) {
			row = i;
			break;
		}
	}

	if(row == -1) {
		row = numberOfRows;
		RGBATableWidget->setRowCount(numberOfRows + 1);
	}

	QTableWidgetItem *itemIntensity =  new QTableWidgetItem();
	itemIntensity->setData(Qt::DisplayRole, intensity);
	RGBATableWidget->setItem(row, 0, itemIntensity);

	QTableWidgetItem *itemR =  new QTableWidgetItem(QString::number(R));
	RGBATableWidget->setItem(row, 1, itemR);

	QTableWidgetItem *itemG =  new QTableWidgetItem(QString::number(G));
	RGBATableWidget->setItem(row, 2, itemG);

	QTableWidgetItem *itemB =  new QTableWidgetItem(QString::number(B));
	RGBATableWidget->setItem(row, 3, itemB);

	QTableWidgetItem *itemA =  new QTableWidgetItem(QString::number(A));
	RGBATableWidget->setItem(row, 4, itemA);

	RGBATableWidget->sortItems(0, Qt::AscendingOrder);
}

int Render::loadAnalyzeImage(const QString &fileName)
{
	vtkSmartPointer<vtkmsqAnalyzeReader> imageReader =
			vtkSmartPointer<vtkmsqAnalyzeReader>::New();

	vtkSmartPointer<vtkEventQtSlotConnect> connection =
			vtkSmartPointer<vtkEventQtSlotConnect>::New();

	if (imageReader->CanReadFile(fileName.toLocal8Bit().constData()) == 0) {
		return 0;
	}

	imageReader->SetFileName(fileName.toLocal8Bit().constData());
	imageReader->SetImageProperties(properties);
	imageReader->UpdateWholeExtent();

	image = vtkImageData::New();
	image->ShallowCopy(imageReader->GetOutput());

	double range[2];
	image->GetPointData()->GetScalars()->GetRange(range);
	int min = (int) range[0];
	int max = (int) range[1];

	configureButtons(min, max);
	renderer = vtkRenderer::New();
	vtkRenderWindow *renderWindow = declarativeView->GetRenderWindow();

	vtkInteractorStyleTrackballCamera *interactorStyle = vtkInteractorStyleTrackballCamera::New();
	vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::New();

	opacityTransferFunction = vtkPiecewiseFunction::New();

  colorTransferFunction = vtkColorTransferFunction::New();

	opacityTransferFunction = vtkPiecewiseFunction::New();
	opacityTransferFunction->AddPoint(min, 0.0);
	opacityTransferFunction->AddPoint(max, 1.0);

	colorTransferFunction = vtkColorTransferFunction::New();
	colorTransferFunction->AddRGBPoint(min, 1.0, 1.0, 1.0);
	colorTransferFunction->AddRGBPoint(max, 0.0, 0.0, 0.0);

	insertRow(min, 255, 255, 255, 0);
	insertRow(max, 0, 0, 0, 100);
	
	vtkVolumeProperty *volumeProperty = vtkVolumeProperty::New();
  volumeProperty->SetColor(colorTransferFunction);
  volumeProperty->SetScalarOpacity(opacityTransferFunction);
  volumeProperty->ShadeOn();
  volumeProperty->SetInterpolationTypeToLinear();

  vtkVolumeRayCastMIPFunction *mipFunction = vtkVolumeRayCastMIPFunction::New();

  vtkVolumeRayCastMapper *volumeMapper = vtkVolumeRayCastMapper::New();
  volumeMapper->SetVolumeRayCastFunction(mipFunction);

  volumeMapper->SetInputConnection(imageReader->GetOutputPort());

  vtkVolume *volume = vtkVolume::New();
  volume->SetMapper(volumeMapper);
  volume->SetProperty(volumeProperty);

  renderer->AddVolume(volume);

	renderer->SetBackground(1.0, 1.0, 1.0);
	WhiteButton->setChecked(true);

	interactor->SetRenderWindow(renderWindow);
	interactor->SetInteractorStyle(interactorStyle);
	interactor->Initialize();
	renderWindow->AddRenderer(renderer);

	declarativeView->update();

	return 1;
}

Render::~Render()
{
	if (properties != NULL) {
		properties->Delete();
	}
}

